/*
 * $Id: main.cpp,v 1.14 2005/07/04 07:49:11 tsuruoka Exp $
 */

#include <stdio.h>
#include <iostream>
#include <string>
#include <map>
#include <set>
#include <list>
#include <fstream>
#include <sstream>
#include <algorithm>
//#include <hash_map>
#include <sys/time.h>
#include <vector>
#include "maxent.h"
#include <ext/hash_map>

using namespace std;

const string MODEL_FILE_DIR = "./model/";

const int MAX_NUM_NODES = 1000;
//const int NUM_TRAINING_SENTENCES = 1000;
//const int NUM_TRAINING_SENTENCES = 10000;
const int NUM_TRAINING_SENTENCES = 9999999;
const double NAIVE_BAYES_FILTERING_PROB = 0.0001;
//const double PARSING_THRESHOLD = 0.3;
double UNCERTAINTY_MARGIN = 0.2;
int MAX_NUM_BRANCH = 1;
const int MAX_NUM_PARSE_NODES = 100;
const double CHUNK_BONUS = 1.0;

const bool KEEP_NP_TMP = false;

int percolate_head(const string & lhs, const vector<string> & rhs);

struct hashfun_str
{
  size_t operator()(const std::string& s) const {
    assert(sizeof(int) == 4 && sizeof(char) == 1);
    const int* p = reinterpret_cast<const int*>(s.c_str());
    size_t v = 0;
    int n = s.size() / 4;
    for (int i = 0; i < n; i++, p++) {
      //      v ^= *p;
      v ^= *p << (4 * (i % 2)); // note) 0 <= char < 128
    }
    int m = s.size() % 4;
    for (int i = 0; i < m; i++) {
      v ^= s[4 * n + i] << (i * 8);
    }
    return v;
  }
};
typedef __gnu_cxx::hash_multimap<string, string, hashfun_str> multimap_type;
//typedef multimap<string, string> multimap_type;

static string discard_suffix(const string & s)
{
  if (KEEP_NP_TMP) { if (s == "NP-TMP") return s; }

  if (s[0] == '-') return s;
  
  for (int i = 1; i < s.size(); i++) {
    if (s[i] == '-') return s.substr(0, i);
    if (s[i] == '=') return s.substr(0, i);
  }
  
  return s;
}

struct NaiveBayes
{
  typedef map<string, int> map_type;
  //  typedef __gnu_cxx::hash_map<string, int, hashfun_str> map_type;
  
  map_type left_count[2];  // 0 for false, 1 for true
  map_type right_count[2];
  map_type rhs_count[2];
  int total[2];
  NaiveBayes() {
    total[0] = total[1] = 0;
  }
  double Prob(const string & left, const string & right, const string & rhs) const {
    const static double K = 2.0;
    int tl = 0, tr = 0, trhs = 0, fl = 0, fr = 0, frhs = 0;
    map_type::const_iterator tli = left_count[1].find(left);
    if (tli != left_count[1].end()) tl = tli->second;
    //    if (tli == left_count[1].end()) return 0;
    //    tl = tli->second;
    map_type::const_iterator tri = right_count[1].find(right);
    if (tri != right_count[1].end()) tr = tri->second;
    //    if (tri == right_count[1].end()) return 0;
    //    tr = tri->second;
    map_type::const_iterator trhsi = rhs_count[1].find(rhs);
    if (trhsi != rhs_count[1].end()) trhs = trhsi->second;
    map_type::const_iterator fli = left_count[0].find(left);
    if (fli != left_count[0].end()) fl = fli->second;
    map_type::const_iterator fri = right_count[0].find(right);
    if (fri != right_count[0].end()) fr = fri->second;
    map_type::const_iterator frhsi = rhs_count[0].find(rhs);
    if (frhsi != rhs_count[0].end()) frhs = frhsi->second;
    const int ftotal = total[0];
    const int ttotal = total[1];
    double t = (tl+1.0) * (tr+1.0) * (trhs+1.0) / ( (ttotal + K) * (ttotal + K));
    double f = (fl+1.0) * (fr+1.0) * (frhs+1.0) / ( (ftotal + K) * (ftotal + K));
    
    //    double t = (left_count[1][left]+1.0) * (right_count[1][right]+1.0) * (rhs_count[1][rhs]+1.0) / ( (total[1]+K) * (total[1]+K));
    //    double f = (left_count[0][left]+1.0) * (right_count[0][right]+1.0) * (rhs_count[0][rhs]+1.0) / ( (total[0]+K) * (total[0]+K));
    return t / (t+f);
  }
  void Add(const int c, const string & left, const string & right, const string & rhs) {
    left_count[c][left]++;
    right_count[c][right]++;
    rhs_count[c][rhs]++;
    total[c]++;
  }
  void Save(const string & filename) const {
    FILE * fp = fopen(filename.c_str(), "w");
    fprintf(fp, "%d\n%d\n", total[0], total[1]);
    for (int i = 0; i < 2; i++) {
      for (map_type::const_iterator j = left_count[i].begin(); j != left_count[i].end(); j++) {
        fprintf(fp, "%d\nleft\n%s\n%d\n", i, j->first.c_str(), j->second);
      }
      for (map_type::const_iterator j = right_count[i].begin(); j != right_count[i].end(); j++) {
        fprintf(fp, "%d\nright\n%s\n%d\n", i, j->first.c_str(), j->second);
      }
      for (map_type::const_iterator j = rhs_count[i].begin(); j != rhs_count[i].end(); j++) {
        fprintf(fp, "%d\nrhs\n%s\n%d\n", i, j->first.c_str(), j->second);
      }
    }
    fclose(fp);
  }
  void Load(const string & filename) {
    assert(total[0] == 0);
    ifstream ifile(filename.c_str());
    if (!ifile) { cerr << "cannot open " << filename << endl; exit(1); }
    string line1, line2;
    getline(ifile, line1);
    getline(ifile, line2);
    sscanf(line1.c_str(), "%d", &total[0]);
    sscanf(line2.c_str(), "%d", &total[1]);
    while (getline(ifile, line1)) {
      string line2, line3, line4;
      getline(ifile, line2);
      getline(ifile, line3);
      getline(ifile, line4);
      int i, n;
      sscanf(line1.c_str(), "%d", &i);
      sscanf(line4.c_str(), "%d", &n);
      if (line2 == "left")       left_count[i][line3] = n;
      else if (line2 == "right") right_count[i][line3] = n;
      else if (line2 == "rhs")   rhs_count[i][line3] = n;
      else { cerr << "load error: " << filename << endl; exit(1); }
    }
    ifile.close();
  }
};

struct NonTerminal
{
  int height;
  string symbol;
  int head;
  int begin;
  int end;
  NonTerminal(const int h, const string & s, const int b, const int e, const int head_ = -1)
    : height(h), symbol(s), begin(b), end(e), head(head_) {};
};

struct Chunk
{
  string symbol;
  int begin;
  int end;
  double probability;
  Chunk(const string & s, int b, int e) : symbol(s), begin(b), end(e), probability(0) {}
  bool operator==(const Chunk & c) const {
    if (begin != c.begin) return false;
    if (end   != c.end) return false;
    return symbol == c.symbol;
  }
  bool operator<(const Chunk & c) const {
    return probability > c.probability;
  }
  void Print() const {
    cout << probability << "\t" << symbol << " " << begin << " " << end << endl;
  }
};

struct NTSequence
{
  vector<NonTerminal> vnt;
  vector<Chunk> chunks;
  vector<string> vw;
  vector<string> vpret;
  int is_base;
  NTSequence() : is_base(false) {}
  NTSequence(const vector<string> & vw_, const vector<string> & vpret_) : vw(vw_), vpret(vpret_), is_base(false) {}
  void Print() const {
    for (int i = 0; i < vnt.size(); i++) {
      int inside = 0, chunk_j = 0;
      for (vector<Chunk>::const_iterator j = chunks.begin(); j != chunks.end(); j++, chunk_j++) {
        if (i >= j->begin && i < j->end) {
          inside = 2;
          if (i == j->begin)
            cout << "(" << chunks[chunk_j].symbol << " ";
          if (i == j->end-1) inside = 3;
          break;
        }
      }
      cout << vnt[i].symbol << "_" << vw[vnt[i].head] << " ";
      if (inside == 3) cout << ") ";
    }
    cout << endl;
  }
};

struct Model
{
  int max_rhs_len;
  multimap_type rhs2lhs;
  set<string> nonterminal_set;
  map<string, ME_Model> detectors;
  map<string, NaiveBayes> filters;
  void Save(const string & filename) const {
    FILE * fp = fopen(filename.c_str(), "w");
    cerr << "saving the model" << endl;
    //    ofstream ofile(filename.c_str());  // takes very long time. why??
    //    ofile << max_rhs_len << endl;
    fprintf(fp, "%d\n", max_rhs_len);
    for (multimap_type::const_iterator i = rhs2lhs.begin(); i != rhs2lhs.end(); i++) {
      fprintf(fp, "%s\n", i->first.c_str());
      fprintf(fp, "%s\n", i->second.c_str());
      //      ofile << i->first << endl;
      //      ofile << i->second << endl;
    }
    fclose(fp);
    //    ofile.close();
    /*
    for (map<string, ME_Model>::const_iterator i = detectors.begin(); i != detectors.end(); i++) {
      cerr << "saving " << i->first << endl;
      i->second.save_to_file("model." + i->first);
    }
    */
  }
  void Load(const string & filename) {
    if (!nonterminal_set.empty()) { cerr << "error: Model::Load()" << endl; exit(1); }
    ifstream ifile(filename.c_str());
    if (!ifile) { cerr << "cannot open " << filename << endl; exit(1); }
    string line;
    getline(ifile, line);
    istringstream is(line);
    is >> max_rhs_len;
    while (getline(ifile, line)) {
      string line2;
      getline(ifile, line2);
      rhs2lhs.insert(make_pair(line, line2));
      nonterminal_set.insert(line2);
    }
    for (set<string>::const_iterator i = nonterminal_set.begin(); i != nonterminal_set.end(); i++) {
      detectors[*i].load_from_file(MODEL_FILE_DIR + "model." + *i);
      filters[*i].Load(MODEL_FILE_DIR + "filter." + *i);
    }
    ifile.close();
  }
};

int percolate_head(const string & lhs, const vector<NonTerminal> & rhs)
{
  vector<string> vs(rhs.size());
  for (int i = 0; i < rhs.size(); i++) vs[i] = rhs[i].symbol;
  return percolate_head(lhs, vs);
}


NonTerminal
calc_height(const char* &p, int & serialno, vector<int> & hv, int & num_tokens,
            vector<string> & vw, vector<string> & vpret)
{
  int org_serialno = serialno;
  if (org_serialno >= hv.size()) { cerr << "error: hv.size()" << endl; exit(1); }
  serialno++;
  
  int c;
  string nonterminal;  
  while (!isspace(c = (*(p++)))) {
    if (c == '(') {
      p--;
      break;
    }
    nonterminal += char(c);
  }
  nonterminal = discard_suffix(nonterminal);
  //  if (nonterminal == "PRT") nonterminal = "ADVP";

  int height = 0;
  vector<NonTerminal> daughters;
  for (;;) {
    while (isspace(c = (*(p++)))) {}

    switch (c) {
    case '(': {
      NonTerminal nt = calc_height(p, serialno, hv, num_tokens, vw, vpret);
      if (nt.height != -1) {
        height = max(height, nt.height);
        daughters.push_back(nt);
      }
      break;
    }
    case ')': {
      if (daughters.size() == 0) return NonTerminal(-1, "", -1, -1);
      hv[org_serialno] = height+1;
      int b = daughters[0].begin;
      int e = daughters[daughters.size()-1].end;
      return NonTerminal(height+1, nonterminal, b, e);
      break;
    }
    default: {
      string terminal;
      terminal += char(c);
      while ( (c = (*(p++))) != ')' ) {
        terminal += char(c);
      }
      if (nonterminal == "-NONE-") return NonTerminal(-1, "", -1, -1);
      string preterminal(nonterminal);
      vpret.push_back(preterminal);
      vw.push_back(terminal);

      hv[org_serialno] = 0;
      int n = num_tokens++;
      return NonTerminal(0, preterminal, n, n+1, n);
    }
    }
  }

}

NonTerminal
parse_ptb(const char* &p, int & serialno, const vector<int> & hv,
          const int height_focused, int depth,
          NTSequence & nts, int & num_tokens,
          bool suppressed = false)
{
  int org_serialno = serialno;
  if (org_serialno >= hv.size()) { cerr << "error: hv.size()" << endl; exit(1); }
  serialno++;

  int c;

  string nonterminal;  
  while (!isspace(c = (*(p++)))) {
    if (c == '(') {
      p--;
      break;
    }
    nonterminal += char(c);
  }
  nonterminal = discard_suffix(nonterminal);
  //  if (nonterminal == "PRT") nonterminal = "ADVP";

  if (depth == 0) nonterminal = "TOP";

  int chunk_begin = -1;
  if (hv[org_serialno] == height_focused) {
    //    cout << "(" << nonterminal << " ";
    chunk_begin = nts.vnt.size();
  }
  
  int height = 0;
  vector<NonTerminal> daughters;
  for (;;) {
    while (isspace(c = (*(p++)))) {}

    switch (c) {
    case '(': {
      bool sup = suppressed;
      int child_serialno = serialno;
      if (hv[child_serialno] < height_focused) sup = true;
      NonTerminal nt = parse_ptb(p, serialno, hv, height_focused, depth + 1, nts, num_tokens, sup);
      if (nt.height != -1) {
        if (!suppressed && hv[child_serialno] < height_focused) {
          if (nt.head == -1) {
            cerr << "error: empty head" << endl; exit(1);
            cout << nt.symbol << " ";
          } else {
            //            cout << nt.symbol << "_" << nt.head << " ";
            nts.vnt.push_back(nt);
          }
        }
        height = max(height, nt.height);
        daughters.push_back(nt);
      }
      break;
    }
    case ')': {
      if (daughters.size() == 0) return NonTerminal(-1, "", -1, -1);
      assert(height+1 == hv[org_serialno]);
      if (hv[org_serialno] == height_focused) {
        //        cout << ") ";
        nts.chunks.push_back(Chunk(nonterminal, chunk_begin, nts.vnt.size()));
      }
      int headn = percolate_head(nonterminal, daughters);
      int b = daughters[0].begin;
      int e = daughters[daughters.size()-1].end;
      if (KEEP_NP_TMP) {
        if (daughters[headn].symbol == "NP-TMP" && nonterminal == "NP")
          nonterminal = "NP-TMP";
      }
      return NonTerminal(height+1, nonterminal, b, e, daughters[headn].head);
      break;
    }
    default: {
      string terminal;
      terminal += char(c);
      while ( (c = (*(p++))) != ')' ) {
        terminal += char(c);
      }
      if (nonterminal == "-NONE-") return NonTerminal(-1, "", -1, -1);
      string preterminal(nonterminal);

      int n = num_tokens++;
      return NonTerminal(0, preterminal, n, n+1, n);
    }
    }
  }
}

bool read_cascaded_chunks(const string & filename)
{
  string line;
  getline(cin, line);
  if (line.size() == 0) return false;
  istringstream is(line);
  string w;
  vector<string> vpret, vw;
  while (is >> w) {
    if (w[0] == '(') continue;
    if (w[0] == ')') continue;
    string::size_type i = w.find_first_of('_');
    string preterm = w.substr(0, i);
    string word    = w.substr(i+1);
    vpret.push_back(preterm);
    vw.push_back(word);
  }
  int n = vw.size();
  vector< list<string> > paren(n+1);
  list<int> lbegin;
  for (int i = 0; i < n+1; i++) lbegin.push_back(i);

  int iter = 0;
  while (1) {
    istringstream is(line);
    int n = 0;
    list<int>::iterator il = lbegin.begin();
    int inside = 0;
    while (is >> w) {
      if (w[0] == '(') {
        paren[*il].push_front(w + " ");
        inside = 1;
        continue;
      }
      if (w[0] == ')') {
        paren[*il].push_back(") ");
        inside = 0;
        continue;
      }
      n++;
      if (inside >= 2) il = lbegin.erase(il);
      else             il++;
      if (inside == 1) inside++;
    }
    //    cout << "iter = " << iter << endl;
    //    for (list<int>::const_iterator j = lbegin.begin(); j != lbegin.end(); j++) {
    //      cout << *j << " ";
    //    }
    //    cout << endl;
    
    getline(cin, line);
    if (line.size() == 0) break;
  }

  for (int i = 0; i < n+1; i++) {
    for (list<string>::const_iterator j = paren[i].begin(); j != paren[i].end(); j++) {
      if ( (*j)[0] == ')') cout << *j;
    }
    for (list<string>::const_iterator j = paren[i].begin(); j != paren[i].end(); j++) {
      if ( (*j)[0] == '(') cout << *j;
    }
    if (i == n) break;
    cout << "(" << vpret[i] << " " << vw[i] << ") ";
  }
  cout << endl;

  
  return true;

}

ME_Sample me_sample(const NTSequence & nts, const Chunk & c, const string & rhs)
{
  ME_Sample mes;

  int n = nts.vnt.size();
  int b = c.begin;
  int e = c.end;

  string Cm1 = "#", Cm2 = "#", Cm3 = "#", Cm4 = "#";
  string CHm1 = "#", CHm2 = "#";
  string CHPm1 = "#", CHPm2 = "#";
  if (b >= 1) {
    Cm1  = nts.vnt[b-1].symbol;
    CHm1 = nts.vw[nts.vnt[b-1].head];
    CHPm1 = nts.vpret[nts.vnt[b-1].head];
  }
  if (b >= 2) {
    Cm2 = nts.vnt[b-2].symbol;
    CHm2 = nts.vw[nts.vnt[b-2].head];
  }
  if (b >= 3) {
    Cm3 = nts.vnt[b-3].symbol;
  }
  if (b >= 4) {
    Cm4 = nts.vnt[b-4].symbol;
  }
  mes.features.push_back("C-1_" + Cm1);
  mes.features.push_back("CH-1_" + CHm1);
  //  if (CHPm1 != Cm1)
  //    mes.features.push_back("CHP-1_" + CHPm1);

  string Cp1 = "#", Cp2 = "#", Cp3 = "#", Cp4 = "#";
  string CHp1 = "#", CHp2 = "#";
  string CHPp1 = "#", CHPp2 = "#";
  if (e < n) {
    Cp1  = nts.vnt[e].symbol;
    CHp1 = nts.vw[nts.vnt[e].head];
    CHPp1 = nts.vpret[nts.vnt[e].head];
  }
  if (e < n-1) {
    Cp2 = nts.vnt[e+1].symbol;
    CHp2 = nts.vw[nts.vnt[e+1].head];
  }
  if (e < n-2) {
    Cp3 = nts.vnt[e+2].symbol;
  }
  if (e < n-3) {
    Cp4 = nts.vnt[e+3].symbol;
  }
  mes.features.push_back("C+1_" + Cp1);
  mes.features.push_back("CH+1_" + CHp1);
  //  if (CHPp1 != Cp1)
  //    mes.features.push_back("CHP+1_" + CHPp1);

  //  string CPm1 = "#", CPm2 = "#";
  //  if (nts.vnt[b].begin >= 1) CPm1 = nts.vpret[nts.vnt[b].begin - 1];
  //  if (nts.vnt[b].begin >= 2) CPm2 = nts.vpret[nts.vnt[b].begin - 2];
  //  mes.features.push_back("CP-1_" + CPm1);

  char buf[1000];
  //  string CPp1 = "#", CPp2 = "#";
  //  if (nts.vnt[e-1].end < nts.vw.size())   CPp1 = nts.vpret[nts.vnt[e-1].end];
  //  if (nts.vnt[e-1].end < nts.vw.size()-1) CPp2 = nts.vpret[nts.vnt[e-1].end+1];
  //  mes.features.push_back("CP+1_" + CPp1);
  int wb = nts.vnt[b].begin;
  int we = nts.vnt[e-1].end;
  string CWm1 = "#", CPm1 = "#";
  if (wb >= 1) {
    CWm1 = nts.vw[b-1];
    CPm1 = nts.vpret[b-1];
  }
  //  mes.features.push_back("CW-1_" + CWm1);
  string CWp1 = "#", CPp1 = "#";
  if (we < nts.vw.size()) {
    CWp1 = nts.vw[e];
    CPp1 = nts.vpret[e];
  }
  //  mes.features.push_back("CW+1_" + CWp1);
  string IWH = nts.vw[b];
  string IWT = nts.vw[e-1];
  string IPH = nts.vpret[b];
  string IPT = nts.vpret[e-1];
  //  mes.features.push_back("IWH_" + IWH);
  //  mes.features.push_back("IWT_" + IWT);
  //  mes.features.push_back("CW-1_IWH_" + CWm1 + "_" + IWH);
  //  mes.features.push_back("CW+1_IWT_" + CWp1 + "_" + IWT);
  //  mes.features.push_back("CP-1_IPH_" + CPm1 + "_" + IPH);
  //  mes.features.push_back("CP+1_IPT_" + CPp1 + "_" + IPT);
    
  //  string str;
  //  for (int i = wb; i < we; i++) str += nts.vw[i] + "_";
  //  mes.features.push_back("STR_" + str);
  //  sprintf(buf, "LEN_L=%d", wb);
  //  mes.features.push_back(buf);
  //  sprintf(buf, "LEN_R=%d", nts.vw.size() - we);
  //  mes.features.push_back(buf);
  //sprintf(buf, "RHS_LEN=%d", rhs.c_str(), we - wb);
  //  mes.features.push_back(buf);

  //  string heads;
  //  for (int i = b; i < e; i++) {
  //    heads += nts.vw[nts.vnt[i].head] + "_";
  //  }
  //  mes.features.push_back("HEADS_" + heads);
  //  mes.features.push_back("RHS_HEADS_" + rhs + "_" + heads);

  //  vector<string> rsyms;
  //  for (int i = b; i < e; i++) rsyms.push_back(nts.vnt[i].symbol);
  //  int headn = percolate_head(c.symbol, rsyms);
  //  string head = nts.vw[nts.vnt[b + headn].head];
  //  mes.features.push_back("HEAD_" + head);
  //  mes.features.push_back("HEAD_CH-1_" + head + "_" + CHm1);
  //  mes.features.push_back("HEAD_CH+1_" + head + "_" + CHp1);

  //  mes.features.push_back("C-4-3-2-1_" + Cm4 + "_" + Cm3 + "_" + Cm2 + "_" + Cm1);
  //  mes.features.push_back("C+1+2+3+4_" + Cp1 + "_" + Cp2 + "_" + Cp3 + "_" + Cp4);
  //  mes.features.push_back("C-3-2-1+1_" + Cm3 + "_" + Cm2 + "_" + Cm1+ "_" + Cp1);
  //  mes.features.push_back("C-2-1+1+2_" + Cm2 + "_" + Cm1 + "_" + Cp1+ "_" + Cp2);
  //  mes.features.push_back("C-1+1+2+3_" + Cm1 + "_" + Cp1 + "_" + Cp2+ "_" + Cp3);

  mes.features.push_back("C-2-1+1_" + Cm2 + "_" + Cm1+ "_" + Cp1);
  mes.features.push_back("C-1+1+2_" + Cm1 + "_" + Cp1+ "_" + Cp2);
  
  mes.features.push_back("C-3-2-1_" + Cm3 + "_" + Cm2+ "_" + Cm1);
  mes.features.push_back("C+1+2+3_" + Cp1 + "_" + Cp2+ "_" + Cp3);
  mes.features.push_back("C-2-1_" + Cm2+ "_" + Cm1);
  mes.features.push_back("C-1+1_" + Cm1+ "_" + Cp1);
  mes.features.push_back("C+1+2_" + Cp1+ "_" + Cp2);

  mes.features.push_back("CH-2-1_" + CHm2+ "_" + CHm1);
  mes.features.push_back("CH+1+2_" + CHp1+ "_" + CHp2);
  mes.features.push_back("CH-1+1_" + CHm1+ "_" + CHp1);
  
  //  mes.features.push_back("SH_C-1" + nts.vnt[b].symbol + "_" + Cm1);
  //  mes.features.push_back("ST_C+1" + nts.vnt[e-1].symbol + "_" + Cp1);

  mes.features.push_back("SHH_" + nts.vnt[b].symbol + "_" + nts.vw[nts.vnt[b].head]);
  mes.features.push_back("SHT_" + nts.vnt[e-1].symbol + "_" + nts.vw[nts.vnt[e-1].head]);
  if (e - b >= 3) {
    for (int i = b + 1; i < e-1; i++) {
      mes.features.push_back("SHM_" + nts.vnt[i].symbol + "_" + nts.vw[nts.vnt[i].head]);
    }
  }
  mes.features.push_back("SBE_" + nts.vnt[b].symbol + "_" + nts.vnt[e-1].symbol);

  // +0.3%
  //  mes.features.push_back("SH_" + nts.vnt[b].symbol);
  //  mes.features.push_back("ST_" + nts.vnt[e-1].symbol);
  //  if (e - b >= 3) {
  //    for (int i = b + 1; i < e-1; i++) {
  //      mes.features.push_back("SM_" + nts.vnt[i].symbol + "_");
  //    }
  //  }

  //  if (e - b >= 2) {
  //    mes.features.push_back("SH_RHS_" + nts.vnt[b].symbol + "_" + nts.vw[nts.vnt[b].head] + "_" + rhs);
  //    mes.features.push_back("ST_RHS_" + nts.vnt[e-1].symbol + "_" + nts.vw[nts.vnt[e-1].head] + "_" + rhs);
  //  }
  //  mes.features.push_back("SH_CH-1" + nts.vnt[b].symbol + "_" + nts.vw[nts.vnt[b].head] + "_" + CHm1);
  //  mes.features.push_back("SH_CH+1" + nts.vnt[e-1].symbol + "_" + nts.vw[nts.vnt[e-1].head] + "_" + CHp1);
  //  if (e - b >= 2) {
  //    mes.features.push_back("SH12_" + nts.vnt[b].symbol + "_" + nts.vw[nts.vnt[b].head] +
  //                           nts.vnt[b+1].symbol + "_" + nts.vw[nts.vnt[b+1].head]);
  //    mes.features.push_back("ST12_" + nts.vnt[e-1].symbol + "_" + nts.vw[nts.vnt[e-1].head] +
  //                           nts.vnt[e-2].symbol + "_" + nts.vw[nts.vnt[e-2].head]);
  //  }
  //  mes.features.push_back("HH_" + nts.vw[nts.vnt[b].head]);
  //  mes.features.push_back("HT_" + nts.vw[nts.vnt[e-1].head]);
  //  mes.features.push_back("SH_" + nts.vnt[b].symbol);
  //  mes.features.push_back("ST_" + nts.vnt[e-1].symbol);
  
  mes.features.push_back("RHS_" + rhs);
  mes.features.push_back("RHS_C-1_" + rhs + "_" + Cm1);
  mes.features.push_back("RHS_C+1_" + rhs + "_" + Cp1);
  mes.features.push_back("RHS_C-1+1_" + rhs + "_" + Cm1 + "_" + Cp1);
  //  mes.features.push_back("RHS_C-2-1_" + rhs + "_" + Cm2 + "_" + Cm1);
  //  mes.features.push_back("RHS_C+1+2_" + rhs + "_" + Cp1 + "_" + Cp2);

  //  int len = nts.vnt[e-1].end - nts.vnt[b].begin;
  //  if (len == e - b) mes.features.push_back("BASE");
  //  if (len == 1) mes.features.push_back("UNARY_" + nts.vw[nts.vnt[b].head]);
  //sprintf(buf, "LEN=%d_%s", len, rhs.c_str());
  //  mes.features.push_back(buf);

  //  mes.features.push_back("HEAD_" + rhs);
  /*
  cout << "chunk = " << c.symbol << " " << c.begin << " " << c.end << endl;
  nts.Print();
  for (vector<string>::const_iterator i = mes.features.begin(); i != mes.features.end(); i++) {
    cout << *i << " ";
  }
  cout << endl << endl;
  */
  return mes;  
}

struct NTNormalizer
{
  multimap_type table;
  NTNormalizer() {
    //    return;
    table.insert(make_pair("VBD", "VBP"));
    table.insert(make_pair("VBZ", "VBP"));
    table.insert(make_pair("JJR", "JJ"));
    table.insert(make_pair("JJS", "JJ"));
    table.insert(make_pair("RBR", "RB"));
    table.insert(make_pair("RBS", "RB"));
    table.insert(make_pair("NNS", "NN"));
    table.insert(make_pair("NNP", "NN"));
    table.insert(make_pair("NNPS", "NN"));
    table.insert(make_pair("PRP", "NN"));
    table.insert(make_pair("PRP$", "JJ"));
    table.insert(make_pair(":", ","));
    
    table.insert(make_pair("''", ""));
    table.insert(make_pair("``", ""));
    
    //    table.insert(make_pair(",", ""));
    //    table.insert(make_pair(":", ""));
    //    table.insert(make_pair(".", ""));
  }
  string Normalize(const string & s) {
    multimap_type::const_iterator i = table.find(s);
    //    if (i == table.end()) return s + "/";
    if (i == table.end()) return s;
    return i->second;
  }
  void update_rhs(vector<string> & rhs, const string & s0)
  {
    //          if (j > 1 && nts.vnt[j-1].symbol == "CC" && nts.vnt[j-2].symbol == nts.vnt[j].symbol) { rhs2 = rhs2.substr(0, rhs2.size()-3); continue; }
    string s = Normalize(s0);
    int n = rhs.size();
    if (n >= 2 && rhs[n-1] == "CC") {
      if (s == rhs[n-2]) { rhs.pop_back(); return; }
      if (n >= 3 && s == rhs[n-3] && rhs[n-2] == ",") {
        rhs.pop_back();
        rhs.pop_back();
        return;
      }
    }
    if (s == "") return;
    if (n > 0 && s == rhs[n-1]) {
      if (s == "NN") return;
      if (s == "JJ") return;
      if (s == "CD") return;
    }
    if (n > 1 && s == rhs[n-2] && rhs[n-1] == ",") {
      rhs.pop_back();
      return;
    }
    rhs.push_back(s);
  }
  string vec2str(const vector<string> & vs)
  {
    string s;
    for (int i = 0; i < vs.size(); i++) s += vs[i] + " ";
    return s;
  }
};

NTNormalizer ntn;

/****
void read_ptb_file(const string & filename, vector<NTSequence> & vnts)
{
  FILE *fp;
  if (!(fp = fopen(filename.c_str(), "r"))) {
    cout << "error: cannot open " << filename << "!" << endl;
    exit(1);
  }

  vector<NTSequence> vnts;

  int c;
  int num_sentences = 0, new_rules = 0;
  max_rhs_len = 0;
  for (;;) {
    cout << "n = " << num_sentences+1 << endl;
    // read one sentence
    int paren = 0;
    string s;
    while (c = fgetc(fp)) {
      if (c == EOF) break;
      s += char(c);
      if (c == '(') paren++;
      if (c == ')') {
        paren--;
        if (paren == 0) break;
      }
    }
    if (c == EOF) break;

    const char *p = s.c_str();
    while (isspace(c = (*(p++)))) {}
    int serialno = 0, num_tokens = 0;
    vector<int> hv(MAX_NUM_NODES);
    vector<string> vw, vpret;
    calc_height(p, serialno, hv, num_tokens, vw, vpret);
    num_sentences++;

    for (int h = 1; h <= hv[0]; h++) {
      p = s.c_str(); serialno = 0;
      while (isspace(c = (*(p++)))) {}
      NTSequence nts(vw, vpret);
      int num_tokens = 0;
      parse_ptb(p, serialno, hv, h, 0, nts, num_tokens);
      if (h == 1) nts.is_base = 1;
      else        nts.is_base = 0;
      vnts.push_back(nts);
      nts.Print();
    }
    cout << endl;
  }

  fclose(fp);
}
***/

void read_sentences(const string & filename, Model & m)
{
  int & max_rhs_len = m.max_rhs_len;
  multimap_type & rhs2lhs = m.rhs2lhs;
  set<string> & nonterminal_set = m.nonterminal_set;
  //  map<string, ME_Model> & detectors = m.detectors;
  
  FILE *fp;
  if (!(fp = fopen(filename.c_str(), "r"))) {
    cout << "error: cannot open " << filename << "!" << endl;
    exit(1);
  }

  vector<NTSequence> vnts;

  int c;
  int num_sentences = 0, new_rules = 0;
  max_rhs_len = 0;
  for (;;) {
    cout << "n = " << num_sentences+1 << endl;
    // read one sentence
    int paren = 0;
    string s;
    while (c = fgetc(fp)) {
      if (c == EOF) break;
      s += char(c);
      if (c == '(') paren++;
      if (c == ')') {
        paren--;
        if (paren == 0) break;
      }
    }
    if (c == EOF) break;
    //    cout << s << endl;

    const char *p = s.c_str();
    while (isspace(c = (*(p++)))) {}
    int serialno = 0, num_tokens = 0;
    vector<int> hv(MAX_NUM_NODES);
    vector<string> vw, vpret;
    calc_height(p, serialno, hv, num_tokens, vw, vpret);
    //    cout << "num_nodes = " << serialno << endl;
    //    cout << "height = " << hv[0] << endl;
    num_sentences++;

    for (int h = 1; h <= hv[0]; h++) {
      p = s.c_str(); serialno = 0;
      while (isspace(c = (*(p++)))) {}
      NTSequence nts(vw, vpret);
      int num_tokens = 0;
      parse_ptb(p, serialno, hv, h, 0, nts, num_tokens);
      if (h == 1) nts.is_base = 1;
      else        nts.is_base = 0;
      vnts.push_back(nts);
      for (vector<Chunk>::const_iterator i = nts.chunks.begin(); i != nts.chunks.end(); i++) {
        const string & lhs = i->symbol;
        nonterminal_set.insert(lhs);
        vector<string> rhs2;
        for (int j = i->begin; j < i->end; j++) {
          ntn.update_rhs(rhs2, nts.vnt[j].symbol);
        }
        if (i->end - i->begin < 30)  // to be addressed later
          max_rhs_len = max(max_rhs_len, i->end - i->begin);
        bool found = false;
        string rhs2str = ntn.vec2str(rhs2);
        pair<multimap_type::const_iterator, multimap_type::const_iterator> itp = rhs2lhs.equal_range(rhs2str);
        for (multimap_type::const_iterator k = itp.first; k != itp.second; k++) {
          if (k->second == lhs) { found = true; break; }
        }
        if (!found) {
          rhs2lhs.insert(make_pair(rhs2str, lhs));
          cout << "***" << lhs << " <- " << rhs2str << endl;
          new_rules++;
        }
      }
      nts.Print();
    }
    cout << endl;
    if (num_sentences % 100 == 0) {
      cout << "the number of newly introduces rules = " << new_rules << endl;
      new_rules = 0;
    }
    if (num_sentences > NUM_TRAINING_SENTENCES) break;
    //    }
    //    exit(0);
    //    print_sentence(lt);
  }

  cerr << "number of symbols = " << nonterminal_set.size() << endl;
  cerr << "max_rhs_len = " << max_rhs_len << endl;

  // print the rule dictionary 
  //  for (multimap_type::const_iterator i = rhs2lhs.begin(); i != rhs2lhs.end(); i++) {
  //    cout << i->second << " <- " << i->first << endl;
  //  }

  for (set<string>::const_iterator ii = nonterminal_set.begin(); ii != nonterminal_set.end(); ii++) {
    const string dname = *ii;
    cerr << "dname = " << dname << endl;

    NaiveBayes nb;
    // building a naive Bayes filter
    for (vector<NTSequence>::const_iterator it = vnts.begin(); it != vnts.end(); it++) {
      const NTSequence & nts = *it;
      const int n = nts.vnt.size();
      for (int len = 1; len <= n; len++) {
        if (len > max_rhs_len) continue;
        for (int i = 0; i < n - len + 1; i++) {
          int begin = i, end = i + len;
          string left = "#", right = "#";
          if (begin >= 1) left = nts.vnt[begin-1].symbol;
          if (end < n)   right = nts.vnt[end].symbol;
          string rhs;
          vector<string> rhs2;
          for (int j = begin; j != end; j++) {
            rhs += nts.vnt[j].symbol + " ";
            ntn.update_rhs(rhs2, nts.vnt[j].symbol);
          }
          string rhs2str = ntn.vec2str(rhs2);
          pair<multimap_type::const_iterator, multimap_type::const_iterator> itp = rhs2lhs.equal_range(rhs2str);
          for (multimap_type::const_iterator j = itp.first; j != itp.second; j++) {
            const string & lhs = j->second;
            if (lhs != *ii) continue;
            Chunk c(lhs, begin, end);
            if (find(nts.chunks.begin(), nts.chunks.end(), c) == nts.chunks.end()) {
              nb.Add(0, left, right, rhs2str);
            } else {
              nb.Add(1, left, right, rhs2str);
            }
          }
        }
      }
    }
    nb.Save(MODEL_FILE_DIR + "filter." + dname);
    
    int num_positives = 0, filtering_error = 0, num_total = 0;
    vector<ME_Sample> train_sets;
    for (vector<NTSequence>::const_iterator it = vnts.begin(); it != vnts.end(); it++) {
      const NTSequence & nts = *it;
      const int n = nts.vnt.size();
      for (int len = 1; len <= n; len++) {
        if (len > max_rhs_len) continue;
        for (int i = 0; i < n - len + 1; i++) {
          int begin = i, end = i + len;
          string left = "#", right = "#";
          if (begin >= 1) left = nts.vnt[begin-1].symbol;
          if (end < n)   right = nts.vnt[end].symbol;
          string rhs;
          vector<string> rhs2;
          for (int j = begin; j != end; j++) {
            rhs += nts.vnt[j].symbol + " ";
            ntn.update_rhs(rhs2, nts.vnt[j].symbol);
          }
          string rhs2str = ntn.vec2str(rhs2);
          pair<multimap_type::const_iterator, multimap_type::const_iterator> itp = rhs2lhs.equal_range(rhs2str);
          for (multimap_type::const_iterator j = itp.first; j != itp.second; j++) {
            const string & lhs = j->second;
            if (lhs != *ii) continue;
            Chunk c(lhs, begin, end);
            num_total++;
            if (nb.Prob(left, right, rhs2str) < NAIVE_BAYES_FILTERING_PROB) {
              if (find(nts.chunks.begin(), nts.chunks.end(), c) != nts.chunks.end()) filtering_error++;
              continue;
            }
            ME_Sample mes = me_sample(nts, c, rhs);
            if (find(nts.chunks.begin(), nts.chunks.end(), c) == nts.chunks.end()) {
              mes.label = "0";
            } else {
              mes.label = "1";
              num_positives++;
            }
            train_sets.push_back(mes);
          }
        }
      }
    }
    cerr << "filtering ratio: " << train_sets.size() << " / " << num_total << " = " << train_sets.size() / (double)num_total << endl;
    cerr << "num_positives = " << num_positives << " missed = " << filtering_error << endl;
    ME_Model model;
    model.set_heldout(train_sets.size() / 10, 0);
    model.train(train_sets, 0, 0, 1);
    model.save_to_file(MODEL_FILE_DIR + "model." + dname);
  }


  fclose(fp);
}

struct ChunkSet
{
  double score;
  vector<Chunk> vc;
  ChunkSet() : score(1) {}
  void Print() const {
    for (vector<Chunk>::const_iterator i = vc.begin(); i != vc.end(); i++) {
      cout << "[" << i->symbol << " " << i->begin << " " << i->end << " " << i->probability << "] ";
    }
    cout << endl;
  }
  bool operator<(const ChunkSet & x) const {
    return score > x.score;
  }
};

int generate_valid_chunksets_recur(const int depth, const vector<Chunk> & candidates,
                                   const vector<int> & chunked_org,
                                   const set<int> & history_org, 
                                   vector<ChunkSet> & vcs)
{
  if (depth == candidates.size()) {
    ChunkSet cs;
    for (set<int>::const_iterator i = history_org.begin(); i != history_org.end(); i++) {
      cs.vc.push_back(candidates[*i]);
      cs.score *= (CHUNK_BONUS * candidates[*i].probability); // 1 -> 2, 0.5 -> 1
    }
    if (cs.vc.size() != 0)
      vcs.push_back(cs);
    return 0;
  }

  const Chunk & candidate = candidates[depth];
  
  vector<int> overlap;
  for (int j = candidate.begin; j != candidate.end; j++) {
    if (chunked_org[j] != -1) overlap.push_back(chunked_org[j]);
  }

  // skip this candidate
  if (overlap.size() > 0) {
    generate_valid_chunksets_recur(depth + 1, candidates, chunked_org, history_org, vcs);
  }
  
  vector<int> chunked = chunked_org;
  set<int> history = history_org;
  for (vector<int>::const_iterator i = overlap.begin(); i != overlap.end(); i++) {
    fill(&chunked[candidates[*i].begin], &chunked[candidates[*i].end], -1);
    history.erase(*i);
  }
  fill(&chunked[candidate.begin], &chunked[candidate.end], depth);
  history.insert(depth);
  generate_valid_chunksets_recur(depth + 1, candidates, chunked, history, vcs);

  if (overlap.size() == 0) {
    if (candidate.probability < 0.5 + UNCERTAINTY_MARGIN)
      generate_valid_chunksets_recur(depth + 1, candidates, chunked_org, history_org, vcs);
  }

  return 0;
}

int generate_valid_chunksets(const int n, vector<Chunk> & candidates,
                            vector<ChunkSet> & vcs)
{
  sort(candidates.begin(), candidates.end());
  vector<int> chunked(n, -1);
  set<int> history;
  generate_valid_chunksets_recur(0, candidates, chunked, history, vcs);

  //  cout << "valid chunk sets" << endl;
  //  for (vector<ChunkSet>::const_iterator i = vcs.begin(); i != vcs.end(); i++)
  //    i->Print();
  //  cout << endl;

}

void perform_chunking(const NTSequence & nts, const Model & m, const double threshold,
                      const bool is_base, vector<ChunkSet> & vcs)
{
  assert(nts.chunks.size() == 0 && vcs.size() == 0);
  const int n = nts.vnt.size();
  vector<Chunk> chunks;
  for (int i = 0; i < n; i++) {
    string rhs;
    vector<string> rhs2;
    for (int len = 1; len <= m.max_rhs_len; len++) {
      if (i + len > n) break;
      int j = i + len - 1;
      rhs += nts.vnt[j].symbol + " ";
      ntn.update_rhs(rhs2, nts.vnt[j].symbol);
      int begin = i;
      int end = i + len;
      string left = "#", right = "#";
      if (begin >= 1) left = nts.vnt[begin-1].symbol;
      if (end < n)   right = nts.vnt[end].symbol;
      string rhs2str = ntn.vec2str(rhs2);
      pair<multimap_type::const_iterator, multimap_type::const_iterator> itp = m.rhs2lhs.equal_range(rhs2str);
      for (multimap_type::const_iterator j = itp.first; j != itp.second; j++) {
        const string & lhs = j->second;
        if (len == 1 && lhs == nts.vnt[begin].symbol) continue; // lhs == rhs!
        map<string, NaiveBayes>::const_iterator fi = m.filters.find(lhs);
        assert(fi != m.filters.end());
        if (fi->second.Prob(left, right, rhs2str) < NAIVE_BAYES_FILTERING_PROB) continue;
        Chunk c(lhs, begin, end);
        ME_Sample mes = me_sample(nts, c, rhs);
        string dname = lhs;
        //        if (is_base) dname = "base" + lhs;
        map<string, ME_Model>::const_iterator it = m.detectors.find(dname);
        if (it->second.num_classes() == 0) continue;
        vector<double> vd = it->second.classify(mes);
        double p = vd[it->second.get_class_id("1")];
        //        cout << "candidate: " << p << "\t";
        //        c.Print();
        if (p >= threshold) {
          c.probability = p;
          chunks.push_back(c);
        }
      }
    }
  }
  //  cout << endl << "chunks" << endl;
  //  for (vector<Chunk>::const_iterator i = chunks.begin(); i != chunks.end(); i++)
  //    i->Print();
  //  cout << endl;
  generate_valid_chunksets(n, chunks, vcs);
  //  nts.chunks = vcs[0].vc;
}

double parse_sentence_recur(int & num_nodes,
                            const int depth, const Model & m, const double threshold,
                            const NTSequence & nts, const vector<int> & vbegin,
                            const vector<Chunk> & unary_history,
                            vector<Chunk> & best_chunks)
{
  //  if (depth > 0) return 1;  // base only
  if (num_nodes++ > MAX_NUM_PARSE_NODES) return 0;
  
  if (depth > 100) { cerr << "warning: depth > 100" << endl; return 0;}
  
  bool is_base = (depth == 0) ? true : false;
  vector<ChunkSet> vcs;
  perform_chunking(nts, m, threshold, is_base, vcs);

  // avoid repetitive applications of the same unary rule on the same nonterminal
  for (vector<ChunkSet>::iterator csi = vcs.begin(); csi != vcs.end(); ) {
    if (csi->vc.size() != 1) { csi++; continue;}
    if (find(unary_history.begin(), unary_history.end(), csi->vc[0]) != unary_history.end())
      csi = vcs.erase(csi);
    else csi++;
  }

  //  nts.Print();
  if (vcs.size() == 0) {
    int n = unary_history.size();
    if (unary_history.size() > 0 && unary_history[n-1].symbol == "TOP" && unary_history[n-1].begin == 0 && unary_history[n-1].end == nts.vnt.size())
      return 1000;
    //    cout << depth << " ";
    return 1;
  }
  double max_score = 0;
  vector<Chunk> max_all_chunks;
  //  sort(vcs.begin(), vcs.end());
  //  if (vcs.size() > 1) vcs.erase(vcs.begin() + 1, vcs.end());
  int num_branch = 0;
  for (vector<ChunkSet>::const_iterator csi = vcs.begin(); csi != vcs.end(); csi++) {
    vector<Chunk> all_chunks;
    assert(csi->vc.size() != 0 && csi->score != 0);
    vector<Chunk> uh;
    if (csi->vc.size() == 1 && csi->vc[0].end - csi->vc[0].begin == 1)
      { uh = unary_history; uh.push_back(csi->vc[0]); }
    for (vector<Chunk>::const_iterator j = csi->vc.begin(); j != csi->vc.end(); j++) {
      Chunk c(*j);
      c.begin = vbegin[j->begin];
      c.end  = vbegin[j->end];
      all_chunks.push_back(c);
    }
    NTSequence new_nts(nts.vw, nts.vpret);
    vector<int> new_vbegin;
    for (int i = 0; i < nts.vnt.size(); i++) {
      int inside = 0, chunk_j = 0;
      for (vector<Chunk>::const_iterator j = csi->vc.begin(); j != csi->vc.end(); j++, chunk_j++) {
        if (i >= j->begin && i < j->end) {
          inside = 2;
          if (i == j->begin) inside = 1;
          if (i == j->end-1) inside = 3;
          break;
        }
      }
      if (inside == 0) {
        new_nts.vnt.push_back(nts.vnt[i]);
        new_vbegin.push_back(vbegin[i]);
        continue;
      }
      if (inside != 3) continue;
      const Chunk & c = csi->vc[chunk_j];
      vector<NonTerminal> daughters(&nts.vnt[c.begin], &nts.vnt[c.end]);
      int headn = percolate_head(c.symbol, daughters);
      int b = daughters[0].begin;
      int e = daughters[daughters.size()-1].end;
      string nonterminal = c.symbol;
      if (KEEP_NP_TMP) {
        if (daughters[headn].symbol == "NP-TMP" && nonterminal == "NP")
          nonterminal = "NP-TMP";
      }
      new_nts.vnt.push_back(NonTerminal(0, nonterminal, b, e, daughters[headn].head));
      new_vbegin.push_back(vbegin[c.begin]);
    }
    new_vbegin.push_back(vbegin[vbegin.size()-1]);

    vector<Chunk> new_all_chunks;
    double score = csi->score *
      parse_sentence_recur(num_nodes, depth + 1, m, threshold, new_nts, new_vbegin, uh, new_all_chunks);

    if (score > max_score) {
      max_score = score;
      max_all_chunks = all_chunks;
      //      copy(new_all_chunks.begin(), new_all_chunks.end(), max_all_chunks.end());
      for (vector<Chunk>::const_iterator i = new_all_chunks.begin(); i != new_all_chunks.end(); i++) max_all_chunks.push_back(*i);
      //      cout << "depth = " << depth << " " << max_all_chunks.size()  << endl;
      //      if (max_all_chunks[max_all_chunks.size()-1].symbol == "TOP") { break;}
    }
    if (++num_branch >= MAX_NUM_BRANCH) break;
  }
  best_chunks = max_all_chunks;
  return max_score;
}

int push_stop_watch()
{
  static struct timeval start_time, end_time;
  static bool start = true;
  if (start) {
    gettimeofday(&start_time, NULL);
    start = false;
    return 0;
  }

  gettimeofday(&end_time, NULL);
  int elapse_msec = (end_time.tv_sec - start_time.tv_sec) * 1000 +
    (int)((end_time.tv_usec - start_time.tv_usec) * 0.001);
  cout << elapse_msec << " msec" << endl;
  start = true;
  return elapse_msec;
}

class ParenConverter
{
  map<string, string> ptb2pos;
  map<string, string> pos2ptb;
public:
  ParenConverter() {
    const static char* table[] = {
      "-LRB-", "(",
      "-RRB-", ")",
      "-LSB-", "[",
      "-RSB-", "]",
      "-LCB-", "{",
      "-RCB-", "}", 
      "***", "***", 
    };

    for (int i = 0;; i+=2) {
      if (string(table[i]) == "***") break;
      ptb2pos.insert(make_pair(table[i], table[i+1]));
      pos2ptb.insert(make_pair(table[i+1], table[i]));
    }
  }
  string Ptb2Pos(const string & s) {
    map<string, string>::const_iterator i = ptb2pos.find(s);
    if (i == ptb2pos.end()) return s;
    return i->second;
  }
  string Pos2Ptb(const string & s) {
    map<string, string>::const_iterator i = pos2ptb.find(s);
    if (i == pos2ptb.end()) return s;
    return i->second;
  }
};

static ParenConverter paren_converter;

//bool parse_sentence(const string & in, const Model & m, const double threshold,
bool parse_sentence(const string & in, const Model & m, double threshold,
                    bool output_for_enju,
                    string & result)
{
  istringstream is(in);
  string token;
  vector<string> vpret, vw;
  while (is >> token) {
    string::size_type i = token.find_last_of('/');
    string word    = token.substr(0, i);
    string preterm = token.substr(i+1);
    word    = paren_converter.Pos2Ptb(word);
    preterm = paren_converter.Pos2Ptb(preterm);
    vpret.push_back(preterm);
    vw.push_back(word);
  }
  int n = vw.size();
  NTSequence nts(vw, vpret);
  for (int i = 0; i < n; i++) nts.vnt.push_back(NonTerminal(0, vpret[i], i, i+1, i));
  vector<int> vbegin;
  for (int i = 0; i < n+1; i++) vbegin.push_back(i);
  vector<Chunk> all_chunks;
  bool parse_success = false;
  vector<Chunk> unary_history;
  int num_nodes = 0;
  double score = parse_sentence_recur(num_nodes, 0, m, threshold, nts, vbegin, unary_history, all_chunks);
  //  if (!output_for_enju) cout << "score = " << score << endl;
  
  vector< list<string> > paren(n+1);
  for (vector<Chunk>::const_iterator i = all_chunks.begin(); i != all_chunks.end(); i++) {
    paren[i->begin].push_front("(" + i->symbol + " ");
    paren[i->end].push_back(") ");
  }

  result = "";
  for (int i = 0; i < n+1; i++) {
    for (list<string>::const_iterator j = paren[i].begin(); j != paren[i].end(); j++) {
      if ( (*j)[0] == ')') result += *j;
    }
    for (list<string>::const_iterator j = paren[i].begin(); j != paren[i].end(); j++) {
      if ( (*j)[0] == '(') result += *j;
    }
    if (i == n) break;
    result += "(" + vpret[i] + " " + vw[i] + ") ";
  }

  if (output_for_enju) {
    vector< list<string> > enju_paren(n+1);
    for (vector<Chunk>::const_iterator i = all_chunks.begin(); i != all_chunks.end(); i++) {
      if (i->symbol == "NP" && vpret[i->begin] == "DT")
        enju_paren[i->begin+1].push_front("(" + i->symbol + " ");
      else 
        enju_paren[i->begin].push_front("(" + i->symbol + " ");
      enju_paren[i->end].push_back(") ");
    }
    string s;
    for (int i = 0; i < n+1; i++) {
      for (list<string>::const_iterator j = enju_paren[i].begin(); j != enju_paren[i].end(); j++) {
        if ( (*j)[0] == ')') s += "\\) ";
      }
      for (list<string>::const_iterator j = enju_paren[i].begin(); j != enju_paren[i].end(); j++) {
        if ( (*j)[0] == '(') s += "\\( ";
      }
      if (i == n) break;
      s += vw[i] + "/" + vpret[i] + " ";
    }
    if (s.size() > 0) s = s.substr(0, s.size()-1);
    cout << s << endl;
  }

  return parse_success;
}

void parse_stdin(const Model & m)
{
  string line;
  int n = 1;
  while (getline(cin, line)) {
    string parsed;
    parse_sentence(line, m, 0.5 - UNCERTAINTY_MARGIN, false, parsed);
    cout << parsed << endl;
  }
}

void parse_file(const string & filename, const Model & m)
{
  ifstream ifile(filename.c_str());
  if (!ifile) {
    cerr << "cannot open " << filename << endl;    exit(1);
  }
  ofstream ofile("results");
  if (!ofile) {
    cerr << "cannot open an output file" << endl; exit(1);
  }
  string line;
  int n = 1;
  while (getline(ifile, line)) {
    string parsed;
    push_stop_watch();
    //    for (double th = 0.5; th >= 0.4; th -= 0.1) {
    //      cout << "threshold = " << th << endl;
    //      if (parse_sentence(line, m, th, parsed)) break;
    //    }
    //    bool success = parse_sentence(line, m, PARSING_THRESHOLD, parsed);
    bool success = parse_sentence(line, m, 0.5 - UNCERTAINTY_MARGIN, false, parsed);
    ofile << parsed << endl;
    cout << "n = " << n++ << endl;
    cout << parsed << endl;
    push_stop_watch();
  }
  
  ifile.close();
  ofile.close();
}

int train(const string & filename)
{
  Model m;
  read_sentences(filename, m);
  m.Save(MODEL_FILE_DIR + "ruledic");
}

int main(int argc, char** argv)
{
  for (int i = 1; i <  argc; i++) {
    string v = argv[i];
    if (v == "-s") {
      UNCERTAINTY_MARGIN = 0.3;
      MAX_NUM_BRANCH = 2;
    }
  }
  //  train("train2-21.mrg");
  //  train("test22.mrg");

  Model m;
  cerr << "loading the model...";
  m.Load(MODEL_FILE_DIR + "ruledic");
  cerr << "done" << endl;
  //  parse_file("wsj22.pos", m);
  //  parse_file("wsj23.pos", m);
  //  parse_file("wsj22.mxpost.pos", m);
  //  parse_file("wsj22.bidir.pos", m);
  //  parse_file("wsj23.mxpost.pos", m);
  parse_stdin(m);
  
  //  read_sentences("tmp.mrg");
}




/*
 * $Log: main.cpp,v $
 * Revision 1.14  2005/07/04 07:49:11  tsuruoka
 * MODEL_FILE_DIR
 *
 * Revision 1.13  2005/06/12 13:06:29  tsuruoka
 * add SBE
 * bugfix SHM
 * true, false -> 1, 0
 *
 * Revision 1.12  2005/06/08 06:59:48  tsuruoka
 * avoid cycle
 *
 * Revision 1.11  2005/06/08 05:08:39  tsuruoka
 * add NTNormalizer
 *
 * Revision 1.10  2005/06/06 04:34:25  tsuruoka
 * add UNCERTAINTY_MARGIN
 *
 * Revision 1.9  2005/06/05 14:22:30  tsuruoka
 * add search
 *
 * Revision 1.8  2005/06/05 07:06:15  tsuruoka
 * add feature
 *
 * Revision 1.7  2005/06/02 11:26:50  tsuruoka
 * add NaiveBayes
 *
 * Revision 1.6  2005/06/01 13:14:40  tsuruoka
 * improve the efficiency of chunking
 *
 * Revision 1.5  2005/06/01 13:04:16  tsuruoka
 * use hash_multimap
 *
 * Revision 1.4  2005/06/01 12:37:18  tsuruoka
 * require less memory
 *
 * Revision 1.3  2005/06/01 09:35:41  tsuruoka
 * add m.Save(), Load()
 * add features
 *
 * Revision 1.2  2005/05/31 09:12:41  tsuruoka
 * add features
 *
 * Revision 1.1  2005/05/30 17:10:56  tsuruoka
 * chunk_parser
 *
 */

